{
  Original Name: Twitter_Data.pas
  Copyright: Copyright (c) 2010, Simon J Stuart, All Rights Reserved
  License: Licensed under the MIT License.
  Author: Simon J Stuart
  Author's Website: http://www.lakraven.com
  Description:
    This unit contains the Parser object along with the various data
    types to hold the information retrieved from the Twitter API.
  Author's Notes: (See "Twitter.pas")
  Acknowledgements: (See "Twitter.pas")

  Version: 0.1 OFFICIAL OPEN BETA RELEASE
  --------------------------------------------------------------------------------
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
  --------------------------------------------------------------------------------
  CODE NOTES: (See "Twitter.pas")

  IMPORTANT NOTICE: You are currently looking at an Open BETA version of TTwitter!
                    This code is subject to a LOT of changes, and whilst you may
                    find portions of this code at present to be messy... rest
                    assured that upon final release, it'll all be nice and clean!
}
unit Twitter_Data;

interface

uses
  Dialogs,
  Classes, Graphics, Windows, SysUtils, DateUtils,
  Twitter_Common,
  Twitter_superobject;

type
  TTwitterEntities = class;
  TTwitterURLFactory = class;
  TTwitterHashTagFactory = class;
  TTwitterMentionsFactory = class;
  TTwitterTweet = class;
  TTwitterUser = class;

  {
    Various OnEvent Methods
  }
  TTwitterAsyncTimelineEvent = procedure(const ATimeline: TTwitterTimelineType; const ATweet: TTwitterTweet) of object;

  {
    TTwitterRateLimitData object contains information about our Twitter IP/API Rate Limits
  }
  TTwitterRateLimitData = record
    HourlyLimit: Integer;
    ResetInSeconds: Integer;
    ResetTime: TDateTime;
    RemainingHits: Integer;
  end;

  {
    TTwitterGeoData object simply contains Latitude and Longitude data.
  }
  TTwitterGeoData = record
    HasCoordinates: Boolean; // will be FALSE if no Coordinate and/or Geo data is available
    Lat: Extended;
    Long: Extended;
  end;

  {
    TTwitterRect object holds data for bounding boxes based on TTwitterGeoData
  }
  TTwitterRect = record
    Left,
    Top,
    Right,
    Bottom: Extended
  end;

  TTwitterIndices = array[0..1] of Integer;

  {
    TTwitterPlaceData object contains information on a location (if given)
  }
  TTwitterPlaceData = record
    HasPlace: Boolean; // will be FALSE if no Place data is available
    PlaceType: String;
    CountryCode: String;
    URL: String;
    BoundingBox: TTwitterRect; // Because AFAIK all "bounding_box" data returned by Twitter are Rectangles!
    FullName: String;
    Name: String;
    Country: String;
    ID: String;
  end;

  {
    TTwitterURL object contains information on a specific URL in a Tweet
  }
  TTwitterURL = record
    StartPos: Integer;
    EndPos: Integer;
    URL: String;
    URLExpanded: String;
    DisplayURL: string;
  end;
  TTwitterURLArray = Array of TTwitterURL;

  {
    TTwitterHashTag object contains information on a specific HashTag (e.g. "#TTwitter") in a Tweet
  }
  [SOOnlySpecified]
  TTwitterHashTag = record
    [SOName('indices')]
    Indices: TTwitterIndices;
    [SOName('text')]
    Tag: String;
  public
    property StartPos: Integer read Indices[0];
    property EndPos: Integer read Indices[1];
  end;
  TTwitterHashTagArray = Array of TTwitterHashTag;

  {
    TTwitterMention object contains information on a specific Mention (user e.g. "@LaKraven") in a Tweet
  }
  TTwitterMention = record
    StartPos: Integer;
    EndPos: Integer;
    ScreenName: String;
    Name: String;
    ID: Int64;
  end;
  TTwitterMentionArray = Array of TTwitterMention;

  {
    TTwitterParser object parses the JSON responses from the Twitter API and returns
    a populated valid object.
  }
  TTwitterParser = class(TObject)
  private
    class function AsType<T>(const AObj: ISuperObject; const AType: TSuperType): T; static;
    function HexToColor(sColor: String): TColor;
    function UTCToSystemTime(UTC: TDateTime): TDateTime;
    function ParseDateTime(sdate: String): TDateTime;
    function ParseNodeTweet(const AObj: ISuperObject): TTwitterTweet;
    function ParseNodeUser(const AObj: ISuperObject): TTwitterUser;
    function ParseNodePlace(const AObj: ISuperObject): TTwitterPlaceData;
    function ParseNodeRect(const AObj: ISuperObject): TTwitterRect;
    function ParseNodeGeo(const AObj: ISuperObject; IsGeoJSON: Boolean): TTwitterGeoData;
    function ParseNodeEntities(const AObj: ISuperObject): TTwitterEntities;
    function ParseNodeURLs(const AObj: ISuperObject): TTwitterURLFactory;
    function ParseNodeURL(const AObj: ISuperObject): TTwitterURL;
    function ParseNodeHashTags(const AObj: ISuperObject): TTwitterHashTagFactory;
    function ParseNodeHashTag(const AObj: ISuperObject): TTwitterHashTag;
    function ParseNodeMentions(const AObj: ISuperObject): TTwitterMentionsFactory;
    function ParseNodeMention(const AObj: ISuperObject): TTwitterMention;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure ParseTimeline(const AData: String; const AOnAsyncTimeline: TTwitterAsyncTimelineEvent; const ATimelineType: TTwitterTimelineType);
    function ParseRateLimit(const AData: String): TTwitterRateLimitData;
    function ParseUserInfo(const AData: String): TTwitterUser;
    function ParseTweet(const AData: String): TTwitterTweet;
  end;

  {
    TTwitterEntities object contains information on URLS, HashTags, Mentions
  }
  TTwitterEntities = class(TObject)
  private
    FURLs: TTwitterURLFactory;
    FHashTags: TTwitterHashTagFactory;
    FMentions: TTwitterMentionsFactory;
  public
    class procedure CopyEntities(const ASource: TTwitterEntities; const ADestination: TTwitterEntities);
    destructor Destroy; override;
    property URLs: TTwitterURLFactory read FURLs;
    property HashTags: TTwitterHashTagFactory read FHashTags;
    property Mentions: TTwitterMentionsFactory read FMentions;
  end;

  {
    TTwitterURLFactory object contains information on all URLs in a Tweet
  }
  TTwitterURLFactory = class(TObject)
  private
    FURLs: TTwitterURLArray;
    FTotal: Integer;
    function GetURL(const AIndex: Integer): TTwitterURL;
  public
    class procedure CopyURLFactory(const ASource, ADestination: TTwitterURLFactory);
    destructor Destroy; override;
    procedure AddNew(const AURL: TTwitterURL);
    property URL[const AIndex: Integer]: TTwitterURL read GetURL; default;
    property Count: Integer read FTotal;
  end;

  {
    TTwitterHashTagFactory object contains information on all HashTags in a Tweet
  }
  TTwitterHashTagFactory = class(TObject)
  private
    FHashTags: TTwitterHashTagArray;
    FTotal: Integer;
    function GetHashTag(const AIndex: Integer): TTwitterHashTag;
  public
    class procedure CopyHashTagFactory(const ASource, ADestination: TTwitterHashTagFactory);
    destructor Destroy; override;
    procedure AddNew(const AHashTag: TTwitterHashTag);
    property HashTag[const AIndex: Integer]: TTwitterHashTag read GetHashTag; default;
    property Count: Integer read FTotal;
  end;

  {
    TTwitterMentionsFactory object contains information on all Mentions (users mentioned e.g. "@LaKraven") in a Tweet
  }
  TTwitterMentionsFactory = class(TObject)
  private
    FMentions: TTwitterMentionArray;
    FTotal: Integer;
    function GetMention(const AIndex: Integer): TTwitterMention;
  public
    class procedure CopyMentionsFactory(const ASource, ADestination: TTwitterMentionsFactory);
    destructor Destroy; override;
    procedure AddNew(const AMention: TTwitterMention);
    property Mention[const AIndex: Integer]: TTwitterMEntion read GetMention; default;
    property Count: Integer read FTotal;
  end;

  {
    TTwitterTweet object contains an actual Tweet, along with methods to reply/delete etc.
  }
  TTwitterTweet = class(TObject)
  private
    [SOName('id')]
    FID: Int64;
    [SOName('created_at')]
    FCreated: TDateTime;
    [SOName('text')]
    FMsg: String;
    FSource: String;
    FIsTruncated: Boolean;
    FReplyTo_TweetID: Int64;
    FReplyTo_UserID: Int64;
    FReplyTo_ScreenName: String;
    FFavorited: Boolean;
    FGeo: TTwitterGeoData;
    FCoordinates: TTwitterGeoData;
    FUser: TTwitterUser;
    FRetweeted: Boolean;
    FRetweetCount: Int64;
    FPlace: TTwitterPlaceData;
    // TODO: NEED TO HANDLE <CONTRIBUTORS/>
    [SOName('entities')]
    FEntities: TTwitterEntities;
  public
    class procedure CopyTweet(const ASource: TTwitterTweet; const ADestination: TTwitterTweet);
    destructor Destroy; override;
    property ID: Int64 read FID;
    property Created: TDateTime read FCreated;
    property Msg: String read FMsg;
    property Source: String read FSource;
    property IsTruncated: Boolean read FIsTruncated;
    property ReplyTo_TweetID: Int64 read FReplyTo_TweetID;
    property ReplyTo_UserID: Int64 read FReplyTo_UserID;
    property ReplyTo_ScreenName: String read FReplyTo_ScreenName;
    property Favorited: Boolean read FFavorited;
    property Geo: TTwitterGeoData read FGeo; // WILL BE NIL IF NO GEO DATA AVAILABLE
    property Coordinates: TTwitterGeoData read FCoordinates; // WILL BE NIL IF NO GEO DATA AVAILABLE
    property User: TTwitterUser read FUser;
    property Retweeted: Boolean read FRetweeted;
    property RetweetCount: Int64 read FRetweetCount;
    property Entities: TTwitterEntities read FEntities;
  end;
  TTweetArray = Array of TTwitterTweet;

  {
    TTwitterProfileStyle object simply contains the user's display options for their Twitter Profile.
    Basically this is how their account on the Twitter Website looks!
    - Useful for any client or app which wants to automatically theme itself to appeal to the client.
    - Implemented by TTwitterUser object.
  }
  TTwitterProfileStyle = record
    ColorBackground: TColor;
    ColorText: TColor;
    ColorLink: TColor;
    ColorSidebarFill: TColor;
    ColorSidebarBorder: TColor;
    ImageBackground: String;
    HasBackgroundTitle: Boolean;
  end;

  {
    TTwitterUser object contains information about a user, and methods to perform user operations
  }
  TTwitterUser = class(TObject)
  private
    FID: Int64;
    FName: String;
    FScreenName: String;
    FLocation: String;
    FDescription: String;
    FProfileImage: String;
    FURL: String;
    FIsProtected: Boolean;
    FFollowers: Int64;
    FFriends: Int64;
    FProfileStyle: TTwitterProfileStyle;
    FCreated: TDateTime;
    FFavorites: Int64;
    FUTCOffset: Integer;
    FTimeZone: String;
{$IF TWITTER_API_VERSION = 1}
    FNotifications: Boolean;
    FFollowing: Boolean;
{$IFEND}
    FGeoEnabled: Boolean;
    FIsVerified: Boolean;
    FLanguage: String;
    FHasContributors: Boolean;
    FTweetCount: Int64;
    FTweetLatest: TTwitterTweet;
  public
    destructor Destroy; override;
    class procedure CopyUser(const ASource: TTwitterUser; const ADestination: TTwitterUser);
    property ID: Int64 read FID;
    property Name: String read FName;
    property ScreenName: String read FScreenName;
    property Location: String read FLocation;
    property Description: String read FDescription;
    property ProfileImage: String read FProfileImage;
    property URL: String read FURL;
    property IsProtected: Boolean read FIsProtected;
    property Followers: Int64 read FFollowers;
    property Friends: Int64 read FFriends;
    property ProfileStyle: TTwitterProfileStyle read FProfileStyle;
    property Created: TDateTime read FCreated;
    property Favorites: Int64 read FFavorites;
    property UTCOffset: Integer read FUTCOffset;
    property TimeZone: String read FTimeZone;
{$IF TWITTER_API_VERSION = 1}
    property Notifications: Boolean read FNotifications;
    property Following: Boolean read FFollowing;
{$IFEND}
    property GeoEnabled: Boolean read FGeoEnabled;
    property IsVerified: Boolean read FIsVerified;
    property Language: String read FLanguage;
    property HasContributors: Boolean read FHasContributors;
    property TweetCount: Int64 read FTweetCount;
    property TweetLatest: TTwitterTweet read FTweetLatest;
  end;
  TTwitterUsers = Array of TTwitterUser;

implementation

{ TTwitterParser }

constructor TTwitterParser.Create;
begin
  inherited;
end;

destructor TTwitterParser.Destroy;
begin
  inherited;
end;

class function TTwitterParser.AsType<T>(const AObj: ISuperObject; const AType: TSuperType): T;
var
  ctx: TSuperRttiContext;
begin
  if not ObjectIsType(AObj, AType) then Exit;
  ctx := TSuperRttiContext.Create;
  try
    Result := ctx.AsType<T>(AObj);
  finally
    ctx.Free;
  end;
end;

function TTwitterParser.HexToColor(sColor: String): TColor;
begin
  if Length(sColor) = 6 then
  begin
    Result := RGB(StrToInt('$' + Copy(String(sColor), 1, 2)),
                  StrToInt('$' + Copy(String(sColor), 3, 2)),
                  StrToInt('$' + Copy(String(sColor), 5, 2)));
  end
  else
    Result := clNone;
end;

function TTwitterParser.UTCToSystemTime(UTC: TDateTime): TDateTime;
var
  TimeZoneInf: _TIME_ZONE_INFORMATION;
  UTCTime, LocalTime: TSystemTime;
begin
  if GetTimeZoneInformation(TimeZoneInf) < $FFFFFFFF then
  begin
    DatetimetoSystemTime(UTC, UTCTime);
    if SystemTimeToTzSpecificLocalTime(@TimeZoneInf, UTCTime, LocalTime) then
    begin
      Result := SystemTimeToDateTime(LocalTime);
    end
    else
      Result := UTC;
  end
  else
    Result := UTC;
end;

function TTwitterParser.ParseDateTime(sdate: String): TDateTime;
var
  dayOfWeek, smonth, dayInMonth, time, offset, year: String;
  hour, minute, second: string;
  month, i: Integer;
begin
  dayOfWeek := Trim(Copy(sdate, 0, 3));
  smonth := Trim(Copy(sdate, 5, 3));
  dayInMonth := Trim(Copy(sdate, 9, 2));
  offset := Trim(Copy(sdate, 20, 6));
  year := Trim(Copy(sdate, 27, 5));
  time := Trim(Copy(sdate, 11, 9));
  hour := Trim(Copy(time, 1, 2));
  minute := Trim(Copy(time, 4, 2));
  second := Trim(Copy(time, 7, 2));
  month := 1; // avoid warning;
  for i := 1 to 12 do
  begin
    if FormatSettings.ShortMonthNames[i] = smonth then
      month := i
  end;
  Result := EncodeDateTime(StrToInt(year), month, StrToInt(dayInMonth),
    StrToInt(hour), StrToInt(minute), StrToInt(second), 0);
end;

function SafeAsBoolean(Obj: ISuperObject): Boolean;
begin
  if Obj <> nil then
    Result := Obj.AsBoolean
  else
    Result := False;
end;

function SafeAsDouble(Obj: ISuperObject): Double;
begin
  if Obj <> nil then
    Result := Obj.AsDouble
  else
    Result := 0;
end;

function SafeAsInteger(Obj: ISuperObject): Int64;
begin
  if Obj <> nil then
    Result := Obj.AsInteger
  else
    Result := 0;
end;

function SafeAsString(Obj: ISuperObject): String;
begin
  if Obj <> nil then
    Result := Obj.AsString
  else
    Result := '';
end;

function TTwitterParser.ParseNodeEntities(const AObj: ISuperObject): TTwitterEntities;
begin
  Result := nil;
  if not ObjectIsType(AObj, stObject) then Exit;

  Result := TTwitterEntities.Create;
  with Result do
  begin
    FURLs := ParseNodeURLs(AObj.AsObject['urls']);
    FHashTags := ParseNodeHashTags(AObj.AsObject['hashtags']);
    FMentions := ParseNodeMentions(AObj.AsObject['user_mentions']);
  end;
end;

function TTwitterParser.ParseNodeGeo(const AObj: ISuperObject; IsGeoJSON: Boolean): TTwitterGeoData;
var
  coord: ISuperObject;
begin
  Result.HasCoordinates := False;
  if not ObjectIsType(AObj, stObject) then Exit;
  coord := AObj.AsObject['coordinates'];
  if not ObjectIsType(coord, stArray) then Exit;

  with Result do
  begin
    if IsGeoJSON then
    begin
      Long := SafeAsDouble(coord.AsArray[0]);
      Lat := SafeAsDouble(coord.AsArray[1]);
    end
    else begin
      Lat := SafeAsDouble(coord.AsArray[0]);
      Long := SafeAsDouble(coord.AsArray[1]);
    end;
  end;
end;

function TTwitterParser.ParseNodeHashTag(const AObj: ISuperObject): TTwitterHashTag;
begin
  Result := AsType<TTwitterHashTag>(AObj, stObject);
end;

function TTwitterParser.ParseNodeHashTags(const AObj: ISuperObject): TTwitterHashTagFactory;
var
  I: Integer;
  arr: TSuperArray;
begin
  Result := nil;
  if not ObjectIsType(AObj, stArray) then Exit;

  Result := TTwitterHashTagFactory.Create;
  arr := AObj.AsArray;
  for I := 0 to arr.Length-1 do
    Result.AddNew(ParseNodeHashTag(arr[I]));
end;

function TTwitterParser.ParseNodeMention(const AObj: ISuperObject): TTwitterMention;
var
  indices: ISuperObject;
begin
  if not ObjectIsType(AObj, stObject) then Exit;

  with Result do
  begin
    indices := AObj.AsObject['indices'];
    if ObjectIsType(indices, stArray) then
    begin
      StartPos := SafeAsInteger(indices.AsArray[0]);
      EndPos := SafeAsInteger(indices.AsArray[1]);
    end;
    ScreenName := SafeAsString(AObj.AsObject['screen_name']);
    Name := SafeAsString(AObj.AsObject['name']);
    ID := SafeAsInteger(AObj.AsObject['id']);
  end;
end;

function TTwitterParser.ParseNodeMentions(const AObj: ISuperObject): TTwitterMentionsFactory;
var
  I: Integer;
  arr: TSuperArray;
begin
  Result := nil;
  if not ObjectIsType(AObj, stArray) then Exit;

  Result := TTwitterMentionsFactory.Create;
  arr := AObj.AsArray;
  for I := 0 to arr.Length-1 do
    Result.AddNew(ParseNodeMention(arr[I]));
end;

function TTwitterParser.ParseNodePlace(const AObj: ISuperObject): TTwitterPlaceData;
begin
  Result.HasPlace := False;
  if not ObjectIsType(AObj, stObject) then Exit;

  with Result do
  begin
    HasPlace := True;
    PlaceType := SafeAsString(AObj.AsObject['place_type']);
    CountryCode := SafeAsString(AObj.AsObject['country_code']);
    URL := SafeAsString(AObj.AsObject['url']);
    BoundingBox := ParseNodeRect(AObj.AsObject['bounding_box']);
    FullName := SafeAsString(AObj.AsObject['full_name']);
    Name := SafeAsString(AObj.AsObject['name']);
    Country := SafeAsString(AObj.AsObject['country']);
    ID := SafeAsString(AObj.AsObject['id']);
  end;
end;

function TTwitterParser.ParseNodeRect(const AObj: ISuperObject): TTwitterRect;
begin
  with Result do
  begin
    Left := 0;
    Top := 0;
    Right := 0;
    Bottom := 0;
    {
    if AObj = nil then Exit;
    if AObj.SelfType = jsNull then Exit;
    if AObj.Field['coordinates'] = nil then
      Exit;
    if AObj.Field['coordinates'].Field[0] = nil then
      Exit;
    if AObj.Field['coordinates'].Field[0].Field[0] = nil then
      Exit;
    if AObj.Field['coordinates'].Field[0].Field[0].Field[0] = nil then
      Exit;
    if AObj.Field['coordinates'].Field[0].Field[1] = nil then
      Exit;
    if AObj.Field['coordinates'].Field[0].Field[1].Field[0] = nil then
      Exit;
    if AObj.Field['coordinates'].Field[0].Field[2] = nil then
      Exit;
    if AObj.Field['coordinates'].Field[0].Field[2].Field[1] = nil then
      Exit;

    Left := JSONSafe(AObj.Field['coordinates'].Field[0].Field[0].Field[0]);
    Top := JSONSafe(AObj.Field['coordinates'].Field[0].Field[0].Field[1]);
    Right := JSONSafe(AObj.Field['coordinates'].Field[0].Field[1].Field[0]);
    Bottom := JSONSafe(AObj.Field['coordinates'].Field[0].Field[2].Field[1]);
    }
  end;
end;

function TTwitterParser.ParseNodeTweet(const AObj: ISuperObject): TTwitterTweet;
begin
  Result := nil;
  if not ObjectIsType(AObj, stObject) then Exit;

  Result := TTwitterTweet.Create;
  with Result do
  begin
    FID := SafeAsInteger(AObj.AsObject['id']);
    FCreated := UTCToSystemTime(ParseDateTime(SafeAsString(AObj.AsObject['created_at'])));
    FMsg := SafeAsString(AObj.AsObject['text']);
    FSource := SafeAsString(AObj.AsObject['source']);
    FIsTruncated := SafeAsBoolean(AObj.AsObject['truncated']);
    FReplyTo_TweetID := SafeAsInteger(AObj.AsObject['in_reply_to_status_id']);
    FReplyTo_UserID := SafeAsInteger(AObj.AsObject['in_reply_to_user_id']);
    FReplyTo_ScreenName := SafeAsString(AObj.AsObject['in_reply_to_screen_name']);
    FFavorited := SafeAsBoolean(AObj.AsObject['favorited']);
    FGeo := ParseNodeGeo(AObj.AsObject['geo'], False);
    FCoordinates := ParseNodeGeo(AObj.AsObject['coordinates'], True);
    FUser := ParseNodeUser(AObj.AsObject['user']);
    FRetweeted := SafeAsBoolean(AObj.AsObject['retweeted']);
    FRetweetCount := SafeAsInteger(AObj.AsObject['retweet_count']);
    FPlace := ParseNodePlace(AObj.AsObject['place']);
    FEntities := ParseNodeEntities(AObj.AsObject['entities']);
  end;
end;

function TTwitterParser.ParseNodeURL(const AObj: ISuperObject): TTwitterURL;
var
  indices: ISuperObject;
begin
  if not ObjectIsType(AObj, stObject) then Exit;

  with Result do
  begin
    indices := AObj.AsObject['indices'];
    if ObjectIsType(indices, stArray) then
    begin
      StartPos := SafeAsInteger(indices.AsArray[0]);
      EndPos := SafeAsInteger(indices.AsArray[1]);
    end;
    URL := SafeAsString(AObj.AsObject['url']);
    URLExpanded := SafeAsString(AObj.AsObject['expanded_url']);
    DisplayURL := SafeAsString(AObj.AsObject['display_url']);
  end;
end;

function TTwitterParser.ParseNodeURLs(const AObj: ISuperObject): TTwitterURLFactory;
var
  I: Integer;
  arr: TSuperArray;
begin
  Result := nil;
  if not ObjectIsType(AObj, stArray) then Exit;

  Result := TTwitterURLFactory.Create;
  arr := AObj.AsArray;
  for I := 0 to arr.Length-1 do
    Result.AddNew(ParseNodeURL(arr[I]));
end;

function TTwitterParser.ParseNodeUser(const AObj: ISuperObject): TTwitterUser;
begin
  Result := nil;
  if not ObjectIsType(AObj, stObject) then Exit;

  Result := TTwitterUser.Create;
  with Result do
  begin
    FID := SafeAsInteger(AObj.AsObject['id']);
    FName := SafeAsString(AObj.AsObject['name']);
    FScreenName := SafeAsString(AObj.AsObject['screen_name']);
    FLocation := SafeAsString(AObj.AsObject['location']);
    FDescription := SafeAsString(AObj.AsObject['description']);
    FProfileImage := SafeAsString(AObj.AsObject['profile_image_url']);
    FURL := SafeAsString(AObj.AsObject['url']);
    FIsProtected := SafeAsBoolean(AObj.AsObject['protected']);
    FFollowers := SafeAsInteger(AObj.AsObject['followers_count']);
    FFriends := SafeAsInteger(AObj.AsObject['friends_count']);
    FCreated := UTCToSystemTime(ParseDateTime(SafeAsString(AObj.AsObject['created_at'])));
    FFavorites := SafeAsInteger(AObj.AsObject['favourites_count']);
    FUTCOffset := SafeAsInteger(AObj.AsObject['utc_offset']);
    FTimeZone := SafeAsString(AObj.AsObject['time_zone']);
    FNotifications := SafeAsBoolean(AObj.AsObject['notifications']);
    FGeoEnabled := SafeAsBoolean(AObj.AsObject['geo_enabled']);
    FIsVerified := SafeAsBoolean(AObj.AsObject['verified']);
    FFollowing := SafeAsBoolean(AObj.AsObject['following']);
    FLanguage := SafeAsString(AObj.AsObject['lang']);
    FHasContributors := SafeAsBoolean(AObj.AsObject['contributors_enabled']);
    FTweetCount := SafeAsInteger(AObj.AsObject['statuses_count']);
    with FProfileStyle do
    begin
      ColorBackground := HexToColor(SafeAsString(AObj.AsObject['profile_background_color']));
      ColorText := HexToColor(SafeAsString(AObj.AsObject['profile_text_color']));
      ColorLink := HexToColor(SafeAsString(AObj.AsObject['profile_link_color']));
      ColorSidebarFill := HexToColor(SafeAsString(AObj.AsObject['profile_sidebar_fill_color']));
      ColorSidebarBorder := HexToColor(SafeAsString(AObj.AsObject['profile_sidebar_border_color']));
      ImageBackground := SafeAsString(AObj.AsObject['profile_background_image_url']);
      HasBackgroundTitle := SafeAsBoolean(AObj.AsObject['profile_background_tile']);
    end;
    FTweetLatest := ParseNodeTweet(AObj.AsObject['status']);
  end;
end;

procedure TTwitterParser.ParseTimeline(const AData: String;
  const AOnAsyncTimeline: TTwitterAsyncTimelineEvent; const ATimelineType: TTwitterTimelineType);
var
  I: Integer;
  obj: ISuperObject;
  arr: TSuperArray;
  LTweet: TTwitterTweet;
begin
  obj := SO(AData);
  if not ObjectIsType(obj, stArray) then Exit;
  arr := obj.AsArray;
  for I := 0 to arr.Length-1 do
  begin
    LTweet := ParseNodeTweet(arr[I]);
    if Assigned(AOnAsyncTimeline) then
      AOnAsyncTimeline(ATimelineType, LTweet);
  end;
end;

function TTwitterParser.ParseTweet(const AData: String): TTwitterTweet;
begin
  Result := ParseNodeTweet(SO(AData));
end;

function TTwitterParser.ParseRateLimit(const AData: String): TTwitterRateLimitData;
var
  obj: ISuperObject;
begin
  obj := SO(AData);
  if not ObjectIsType(obj, stObject) then Exit;
  with Result do
  begin
    HourlyLimit := SafeAsInteger(obj.AsObject['hourly_limit']);
    ResetTime := UTCToSystemTime(ParseDateTime(SafeAsString(obj.AsObject['reset_time'])));
    ResetInSeconds := Round((ResetTime - Now()) * 86400); // JSON_GrabTier1(JSON, 'reset_time_in_seconds'); // < Gives us an unusually massive number!
    RemainingHits := SafeAsInteger(obj.AsObject['remaining_hits']);
  end;
end;

function TTwitterParser.ParseUserInfo(const AData: String): TTwitterUser;
begin
  Result := ParseNodeUser(SO(AData));
end;

{ TTwitterEntities }

class procedure TTwitterEntities.CopyEntities(const ASource, ADestination: TTwitterEntities);
begin
  if (ASource = nil) or (ADestination = nil) then
    Exit;

  if ASource.FURLs <> nil then
  begin
    ADestination.FURLs := TTwitterURLFactory.Create;
    TTwitterURLFactory.CopyURLFactory(ASource.FURLs, ADestination.FURLs);
  end;

  if ASource.FHashTags <> nil then
  begin
    ADestination.FHashTags := TTwitterHashTagFactory.Create;
    TTwitterHashTagFactory.CopyHashTagFactory(ASource.FHashTags, ADestination.FHashTags);
  end;

  if ASource.FMentions <> nil then
  begin
    ADestination.FMentions := TTwitterMentionsFactory.Create;
    TTwitterMentionsFactory.CopyMentionsFactory(ASource.FMentions, ADestination.FMentions);
  end;

end;

destructor TTwitterEntities.Destroy;
begin
  if FURLs <> nil then
    FreeAndNil(FURLs);
  if FMentions <> nil then
    FreeAndNil(FMentions);
  if FHashTags <> nil then
    FreeAndNil(FHashTags);
  inherited;
end;

{ TTwitterURLFactory }

procedure TTwitterURLFactory.AddNew(const AURL: TTwitterURL);
var
  LIdx: Integer;
begin
  LIdx := Length(FURLs);
  SetLength(FURLs, LIdx + 1);
  FURLs[LIdx] := AURL;
  FTotal := LIdx + 1;
end;

class procedure TTwitterURLFactory.CopyURLFactory(const ASource, ADestination: TTwitterURLFactory);
var
  I: Integer;
begin
  if (ASource = nil) or (ADestination = nil) then
    Exit;
  ADestination.FTotal := ASource.FTotal;
  SetLength(ADestination.FURLs, Length(ASource.FURLs));
  for I := 0 to Length(ASource.FURLs) - 1 do
  begin
    ADestination.FURLs[I] := ASource.FURLs[I];
  end;
end;

destructor TTwitterURLFactory.Destroy;
begin
  SetLength(FURLs, 0);
  inherited;
end;

function TTwitterURLFactory.GetURL(const AIndex: Integer): TTwitterURL;
begin
  Result := FURLs[AIndex];
end;

{ TTwitterHashTagFactory }

procedure TTwitterHashTagFactory.AddNew(const AHashTag: TTwitterHashTag);
var
  LIdx: Integer;
begin
  LIdx := Length(FHashTags);
  SetLength(FHashTags, LIdx + 1);
  FHashTags[LIdx] := AHashTag;
  FTotal := LIdx + 1;
end;

class procedure TTwitterHashTagFactory.CopyHashTagFactory(const ASource, ADestination: TTwitterHashTagFactory);
var
  I: Integer;
begin
  if (ASource = nil) or (ADestination = nil) then
    Exit;
  ADestination.FTotal := ASource.FTotal;
  SetLength(ADestination.FHashTags, Length(ASource.FHashTags));
  for I := 0 to Length(ASource.FHashTags) - 1 do
  begin
    ADestination.FHashTags[I] := ASource.FHashTags[I];
  end;
end;

destructor TTwitterHashTagFactory.Destroy;
begin
  SetLength(FHashTags, 0);
  inherited;
end;

function TTwitterHashTagFactory.GetHashTag(const AIndex: Integer): TTwitterHashTag;
begin
  Result := FHashTags[AIndex];
end;

{ TTwitterMentionsFactory }

procedure TTwitterMentionsFactory.AddNew(const AMention: TTwitterMention);
var
  LIdx: Integer;
begin
  LIdx := Length(FMentions);
  SetLength(FMentions, LIdx + 1);
  FMentions[LIdx] := AMention;
  FTotal := LIdx + 1;
end;

class procedure TTwitterMentionsFactory.CopyMentionsFactory(const ASource, ADestination: TTwitterMentionsFactory);
var
  I: Integer;
begin
  if (ASource = nil) or (ADestination = nil) then
    Exit;
  ADestination.FTotal := ASource.FTotal;
  SetLength(ADestination.FMentions, Length(ASource.FMentions));
  for I := 0 to Length(ASource.FMentions) - 1 do
  begin
    ADestination.FMentions[I] := ASource.FMentions[I];
  end;
end;

destructor TTwitterMentionsFactory.Destroy;
begin
  SetLength(FMentions, 0);
  inherited;
end;

function TTwitterMentionsFactory.GetMention(const AIndex: Integer): TTwitterMention;
begin
  Result := FMentions[AIndex];
end;

{ TTwitterTweet }

class procedure TTwitterTweet.CopyTweet(const ASource: TTwitterTweet; const ADestination: TTwitterTweet);
begin
  if (ASource = nil) or (ADestination = nil) then
    Exit;

  ADestination.FID := ASource.FID;
  ADestination.FCreated := ASource.FCreated;
  ADestination.FMsg := ASource.FMsg;
  ADestination.FSource := ASource.FSource;
  ADestination.FIsTruncated := ASource.FIsTruncated;
  ADestination.FReplyTo_TweetID := ASource.FReplyTo_TweetID;
  ADestination.FReplyTo_UserID := ASource.FReplyTo_UserID;
  ADestination.FReplyTo_ScreenName := ASource.FReplyTo_ScreenName;
  ADestination.FFavorited := ASource.FFavorited;
  ADestination.FGeo := ASource.FGeo;
  ADestination.FCoordinates := ASource.FCoordinates;
  ADestination.FUser := ASource.FUser;
  ADestination.FRetweeted := ASource.FRetweeted;
  ADestination.FRetweetCount := ASource.FRetweetCount;
  ADestination.FPlace := ASource.FPlace;
  if ASource.FEntities <> nil then
  begin
    ADestination.FEntities := TTwitterEntities.Create;
    TTwitterEntities.CopyEntities(ASource.FEntities, ADestination.FEntities);
  end;
end;

destructor TTwitterTweet.Destroy;
begin
  if FUser <> nil then
    FreeAndNil(FUser);
  if FEntities <> nil then
    FreeAndNil(FEntities);
  inherited;
end;

{ TTwitterUser }

class procedure TTwitterUser.CopyUser(const ASource: TTwitterUser; const ADestination: TTwitterUser);
begin
  if (ASource = nil) or (ADestination = nil) then
    Exit;

  ADestination.FID := ASource.FID;
  ADestination.FName := ASource.FName;
  ADestination.FScreenName := ASource.FScreenName;
  ADestination.FLocation := ASource.FLocation;
  ADestination.FDescription := ASource.FDescription;
  ADestination.FProfileImage := ASource.FProfileImage;
  ADestination.FURL := ASource.FURL;
  ADestination.FIsProtected := ASource.FIsProtected;
  ADestination.FFollowers := ASource.FFollowers;
  ADestination.FFriends := ASource.FFriends;
  ADestination.FProfileStyle := ASource.FProfileStyle;
  ADestination.FCreated := ASource.FCreated;
  ADestination.FFavorites := ASource.FFavorites;
  ADestination.FUTCOffset := ASource.FUTCOffset;
  ADestination.FTimeZone := ASource.FTimeZone;
{$IF TWITTER_API_VERSION = 1}
  ADestination.FNotifications := ASource.FNotifications;
  ADestination.FFollowing := ASource.FFollowing;
{$IFEND}
  ADestination.FGeoEnabled := ASource.FGeoEnabled;
  ADestination.FIsVerified := ASource.FIsVerified;
  ADestination.FLanguage := ASource.FLanguage;
  ADestination.FHasContributors := ASource.FHasContributors;
  ADestination.FTweetCount := ASource.FTweetCount;
  if ASource.FTweetLatest <> nil then
  begin
    ADestination.FTweetLatest := TTwitterTweet.Create;
    TTwitterTweet.CopyTweet(ASource.FTweetLatest, ADestination.FTweetLatest);
  end;
end;

destructor TTwitterUser.Destroy;
begin
  if FTweetLatest <> nil then
    FreeAndNil(FTweetLatest);
  inherited;
end;

end.
