{
  Original Name: Twitter_TokenStore.pas
  Copyright: Copyright (c) 2010, Simon J Stuart, All Rights Reserved
  License: Licensed under the MIT License.
  Author: Simon J Stuart
  Author's Website: http://www.lakraven.com
  Description:
    This unit contains everything related to storage/reading of User Tokens.
    It provides several variations, all of which are interchangable, and one
    of which allows you to quickly produce your own Token Storage system to
    suit your needs.
  Author's Notes: (See "Twitter.pas")
  Acknowledgements: (See "Twitter.pas")

  Version: 0.1 OFFICIAL OPEN BETA RELEASE
  --------------------------------------------------------------------------------
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
  --------------------------------------------------------------------------------
  CODE NOTES: (See "Twitter.pas")

  IMPORTANT NOTICE: You are currently looking at an Open BETA version of TTwitter!
                    This code is subject to a LOT of changes, and whilst you may
                    find portions of this code at present to be messy... rest
                    assured that upon final release, it'll all be nice and clean!
}
unit Twitter_TokenStore;

interface

uses
  Classes, SysUtils, Registry, IniFiles, Windows,
  Twitter_Common;

type
  {
    Various OnEvent Methods
  }
  TTwitterTokenEvent = procedure(const AToken: TTwitterUserToken) of object;
  TTwitterGetTokenEvent = procedure(var AToken: TTwitterUserToken) of object;

  {
    Various Enum Types
  }
  TTwitterRegistryRoots = (hkClassesRoot, hkCurrentUser, hkLocalMachine, hkUsers, hkPerformanceData, hkCurrentConfig);

  {
    TTwitterTokenManager object is an abstract class on which our Descendant Token Store types build.
  }
  TTwitterTokenManager = class(TComponent)
  public
    procedure StoreToken(const AUserToken: TTwitterUserToken); virtual; abstract;
    procedure GetToken(var AUserToken: TTwitterUserToken); virtual; abstract;
    procedure KillToken(const AUserToken: TTwitterUserToken); virtual; abstract;
  end;

  {
    TTwitterTokenRegistry object is TTwitter's internal default Token Manager
  }
  TTwitterTokenRegistry = class(TTwitterTokenManager)
  private
    FRootKey: TTwitterRegistryRoots;
    FPath: String;
    procedure CheckPath;
    function GetRegistryData(RootKey: HKEY; Key, Value: string): variant;
    procedure SetRegistryData(RootKey: HKEY; Key, Value: string; RegDataType: TRegDataType; Data: variant);
    procedure RemoveRegistryData(RootKey: HKEY; const Key: String);
  public
    constructor Create(AOwner: TComponent); override;
    procedure StoreToken(const AUserToken: TTwitterUserToken); override;
    procedure GetToken(var AUserToken: TTwitterUserToken); override;
    procedure KillToken(const AUserToken: TTwitterUserToken); override;
  published
    property RootKey: TTwitterRegistryRoots read FRootKey write FRootKey;
    property Path: String read FPath write FPath;
  end;

  {
    TTwitterTokenINIStore object is an INI-based alternative to using the Registry
  }
  TTwitterTokenINIStore = class(TTwitterTokenManager)
  private
    FIniFile: TIniFile;
    FFileName: String;
  public
    constructor Create(AOwner: TComponent); override;
    procedure StoreToken(const AUserToken: TTwitterUserToken); override;
    procedure GetToken(var AUserToken: TTwitterUserToken); override;
    procedure KillToken(const AUserToken: TTwitterUserToken); override;
  published
    property FileName: String read FFileName write FFileName;
  end;

  {
    TTwitterTokenCustomStore object is a custom Token Manager using OnEvent methods
    which allow you to determine your own bespoke method of storing, reading and
    deleting User OAuth Tokens.
  }
  TTwitterTokenCustomStore = class(TTwitterTokenManager)
  private
    FOnStoreToken: TTwitterTokenEvent;
    FOnGetToken: TTwitterGetTokenEvent;
    FOnKillToken: TTwitterTokenEvent;
  public
    procedure StoreToken(const AUserToken: TTwitterUserToken); override;
    procedure GetToken(var AUserToken: TTwitterUserToken); override;
    procedure KillToken(const AUserToken: TTwitterUserToken); override;
  published
    property OnStoreToken: TTwitterTokenEvent read FOnStoreToken write FOnStoreToken;
    property OnGetToken: TTwitterGetTokenEvent read FOnGetToken write FOnGetToken;
    property OnKillToken: TTwitterTokenEvent read FOnKillToken write FOnKillToken;
  end;

const
  TTwitterRegistryRootValues: Array[TTwitterRegistryRoots] of DWORD = (DWORD($80000000), DWORD($80000001), DWORD($80000002), DWORD($80000003), DWORD($80000004), DWORD($80000005));

implementation

{ TTwitterTokenRegistry }

procedure TTwitterTokenRegistry.CheckPath;
begin
  if FPath[1] = '/' then
    FPath := Copy(FPath, 2, Length(FPath));
  if FPath[Length(FPath) -1] = '/' then
    FPath := Copy(FPath, 1, Length(FPath) -1);
end;

constructor TTwitterTokenRegistry.Create(AOwner: TComponent);
begin
  inherited;
  FRootKey := hkCurrentUser;
  Path := 'Software\TTwitter\Stored';
end;

function TTwitterTokenRegistry.GetRegistryData(RootKey: HKEY; Key, Value: string): variant;
var
  Reg: TRegistry;
  RegDataType: TRegDataType;
  DataSize: Integer;
  s: string;
begin
  Reg := TRegistry.Create(KEY_QUERY_VALUE);
  Reg.RootKey := RootKey;
  try
    if Reg.OpenKeyReadOnly(Key) then
    begin
      RegDataType := Reg.GetDataType(Value);
      if (RegDataType = rdString) or (RegDataType = rdExpandString) then
        Result := Reg.ReadString(Value)
      else if RegDataType = rdInteger then
        Result := Reg.ReadInteger(Value)
      else if RegDataType = rdBinary then
      begin
        DataSize := Reg.GetDataSize(Value);
        SetLength(s, DataSize);
        Reg.ReadBinaryData(Value, PChar(s)^, DataSize);
        Result := s;
      end;
    end;
  finally
    s := '';
    Reg.CloseKey;
    Reg.Free;
  end;
end;

procedure TTwitterTokenRegistry.RemoveRegistryData(RootKey: HKEY; const Key: String);
var
  Reg: TRegistry;
begin
  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;
    if Reg.OpenKey(Key, False) then
    begin
      Reg.DeleteKey(Key);
    end;
    Reg.CloseKey;
  finally
    Reg.Free;
  end;
end;

procedure TTwitterTokenRegistry.SetRegistryData(RootKey: HKEY; Key, Value: string; RegDataType: TRegDataType; Data: variant);
var
  Reg: TRegistry;
  s: string;
begin
  Reg := TRegistry.Create(KEY_WRITE);
  try
    Reg.RootKey := RootKey;
    if Reg.OpenKey(Key, True) then
    begin
      try
        if RegDataType = rdUnknown then
          RegDataType := Reg.GetDataType(Value);
        if RegDataType = rdString then
          Reg.WriteString(Value, Data)
        else if RegDataType = rdExpandString then
          Reg.WriteExpandString(Value, Data)
        else if RegDataType = rdInteger then
          Reg.WriteInteger(Value, Data)
        else if RegDataType = rdBinary then
        begin
          s := Data;
          Reg.WriteBinaryData(Value, PChar(s)^, Length(s));
        end;
      finally
        Reg.CloseKey;
      end;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TTwitterTokenRegistry.GetToken(var AUserToken: TTwitterUserToken);
begin
  CheckPath;
  with AUserToken do
  begin
    Token := GetRegistryData(TTwitterRegistryRootValues[FRootKey], '\' + Path + '\' + Username, 'Token');
    TokenSecret := GetRegistryData(TTwitterRegistryRootValues[FRootKey], '\' + Path + '\' + Username, 'TokenSecret');
  end;
end;

procedure TTwitterTokenRegistry.KillToken(const AUserToken: TTwitterUserToken);
begin
  CheckPath;
  with AUserToken do
  begin
    RemoveRegistryData(TTwitterRegistryRootValues[FRootKey], '\' + Path + '\' + Username);
  end;
end;

procedure TTwitterTokenRegistry.StoreToken(const AUserToken: TTwitterUserToken);
begin
  CheckPath;
  with AUserToken do
  begin
    SetRegistryData(TTwitterRegistryRootValues[FRootKey], '\' + Path + '\' + Username, 'Token', rdString, Token);
    SetRegistryData(TTwitterRegistryRootValues[FRootKey], '\' + Path + '\' + Username, 'TokenSecret', rdString, TokenSecret);
  end;
end;

{ TTwitterTokenINIStore }

constructor TTwitterTokenINIStore.Create(AOwner: TComponent);
begin
  inherited;
  FFileName := 'Users.ini';
end;

procedure TTwitterTokenINIStore.GetToken(var AUserToken: TTwitterUserToken);
begin
  FIniFile := TIniFile.Create(FFileName);
  try
    with AUserToken do
    begin
      Token := FIniFile.ReadString(Username, 'Token', '');
      TokenSecret := FIniFile.ReadString(Username, 'TokenSecret', '');
    end;
  finally
    FIniFile.Free;
  end;
end;

procedure TTwitterTokenINIStore.KillToken(const AUserToken: TTwitterUserToken);
begin
  FIniFile := TIniFile.Create(FFileName);
  try
    with AUserToken do
    begin
      FIniFile.DeleteKey(Username, 'Token');
      FIniFile.DeleteKey(Username, 'TokenSecret');
    end;
  finally
    FIniFile.Free;
  end;
end;

procedure TTwitterTokenINIStore.StoreToken(const AUserToken: TTwitterUserToken);
begin
  FIniFile := TIniFile.Create(FFileName);
  try
    with AUserToken do
    begin
      FIniFile.WriteString(Username, 'Token', Token);
      FIniFile.WriteString(Username, 'TokenSecret', TokenSecret);
    end;
  finally
    FIniFile.Free;
  end;
end;

{ TTwitterTokenCustomStore }

procedure TTwitterTokenCustomStore.GetToken(var AUserToken: TTwitterUserToken);
begin
  if Assigned(FOnGetToken) then
    FOnGetToken(AUserToken);
end;

procedure TTwitterTokenCustomStore.KillToken(const AUserToken: TTwitterUserToken);
begin
  if Assigned(FOnKillToken) then
    FOnKillToken(AUserToken);
end;

procedure TTwitterTokenCustomStore.StoreToken(const AUserToken: TTwitterUserToken);
begin
  if Assigned(FOnStoreToken) then
    FOnStoreToken(AUserToken);
end;

end.
